/*
    imbLoader.cpp
    Loader plug-in for Irrlicht engine.

    MUTATE Systems 2018 (https://mutate.jp)
*/

#include "ImbLoader.h"
#include "IrrMeshBuild.h"

static irr::scene::ISkinnedMesh::SJoint* findJoint(const char* jointName, irr::scene::ISkinnedMesh* mesh)
{
    for(unsigned int i = 0; i < mesh->getJointCount(); ++i)
        if(!strcmp(mesh->getJointName(i), jointName))
            return mesh->getAllJoints()[i];
    return NULL;
}

class SkinnedVertex
{
public:
    SkinnedVertex() : Moved(false) {}
    bool Moved;
    irr::core::vector3df Position;
    irr::core::vector3df Normal;
};

static irr::video::ITexture* getTexture(irr::video::IVideoDriver* driver, const char* filename, const char** txDirs) {
    for(int i = 0; txDirs[i]; i ++) {
        const char* szTxDir = txDirs[i];
        std::string path = std::string(txDirs[i]) + "/" + filename;
        auto r = driver->getTexture(path.c_str());
        if(r) return r;
    }
    return NULL;
}

static irr::scene::ISkinnedMesh* getMeshFromIMB(SIrrMeshBuild* pid, irr::scene::ISceneManager* smgr, const char** txDirs) {
    auto mesh = smgr->createSkinnedMesh();
    irr::core::array<SkinnedVertex> skinnedVertex;

    // create nodes
    for(auto i = pid->nodes.begin(); i != pid->nodes.end(); i ++) {
        auto& node = (*i);
        irr::scene::ISkinnedMesh::SJoint* jointParent = 
            node.nameParent.length() ? findJoint(node.nameParent.c_str(), mesh) : NULL;

        auto joint = mesh->addJoint(jointParent);
        joint->Name = node.name.c_str();
        joint->LocalMatrix = node.mtxLocal;
        joint->GlobalMatrix = jointParent ? jointParent->GlobalMatrix * node.mtxLocal : node.mtxLocal;

        for(auto j = node.meshes.begin(); j != node.meshes.end(); j ++)
            joint->AttachedMeshes.push_back((*j));
    }

    //create meshes
    for(auto ib = pid->meshbufs.begin(); ib != pid->meshbufs.end(); ib ++) {
        auto buffer = mesh->addMeshBuffer();

        auto& smat = buffer->Material;
        smat.DiffuseColor  = (*ib).scDiffuse;
        smat.EmissiveColor = (*ib).scEmissive;
        smat.SpecularColor = (*ib).scSpecular;

        smat.Shininess = (irr::f32)floor(((*ib).fShininess - 2.0) / 5.0 * 1024 + 0.5);
        if(smat.Shininess > 1024) smat.Shininess = 1024;

        smat.MaterialType = (*ib).matType;

        if((*ib).txDiffuse.length() > 0) {
            smat.setTexture(0, getTexture(smgr->getVideoDriver(), (*ib).txDiffuse.c_str(), txDirs));
            smat.DiffuseColor.set(0xffffffff);
        }
        smat.AmbientColor = smat.DiffuseColor;

        if((*ib).txNormal.length() > 0) {
            smat.setTexture(0, getTexture(smgr->getVideoDriver(), (*ib).txNormal.c_str(), txDirs));
            smat.MaterialType = irr::video::EMT_PARALLAX_MAP_SOLID;
        }

        for(auto imf = (*ib).matFlags.begin(); imf != (*ib).matFlags.end(); imf ++) {
            smat.setFlag((*imf).emf, (*imf).bVal);
        }

        irr::u32 alpha = 255 - (*ib).scTransparent.getRed();
        if(alpha < 255) buffer->Material.MaterialType = irr::video::EMT_TRANSPARENT_VERTEX_ALPHA;

        const size_t vertexCount = (*ib).verts.size();
        buffer->Vertices_Standard.reallocate(vertexCount);
        buffer->Vertices_Standard.set_used(vertexCount);
        for(size_t vn = 0; vn < vertexCount; vn ++) {
            buffer->Vertices_Standard[vn] = (*ib).verts[vn];
            buffer->Vertices_Standard[vn].Color.setAlpha(
                buffer->Vertices_Standard[vn].Color.getAlpha() * alpha / 255 );
        }
        if((*ib).Tangents.size()) {
            buffer->convertToTangents();
            for(size_t vn = 0; vn < vertexCount; vn ++) {
                buffer->Vertices_Tangents[vn].Tangent = (*ib).Tangents[vn];
                buffer->Vertices_Tangents[vn].Binormal = (*ib).Binormals[vn];
            }
        }
        if((*ib).TCoords2.size()) {
            buffer->convertTo2TCoords();
            for(size_t vn = 0; vn < vertexCount; vn ++)
                buffer->Vertices_2TCoords[vn].TCoords2 = (*ib).TCoords2[vn];
        }

        buffer->Indices.reallocate((*ib).indices.size());
        buffer->Indices.set_used((*ib).indices.size());

        for (unsigned j = 0; j < (*ib).indices.size(); ++j)
            buffer->Indices[j] = (*ib).indices[j];

        buffer->recalculateBoundingBox();

        if(!(*ib).bHasNormal)
            smgr->getMeshManipulator()->recalculateNormals(buffer);

        // Skinning

        skinnedVertex.clear();
        skinnedVertex.reallocate(vertexCount);
        for (unsigned vn = 0; vn < vertexCount; ++vn)
            skinnedVertex.push_back(SkinnedVertex());

        for (auto ibone = (*ib).bones.begin(); ibone != (*ib).bones.end(); ibone ++) {
            auto& bone = (*ibone);
            auto joint = findJoint(bone.name.c_str(), mesh);

            for(auto ibw = bone.weights.begin(); ibw != bone.weights.end(); ibw ++) {
                auto& bw = (*ibw);
                auto w = mesh->addWeight(joint);
                w->buffer_id = mesh->getMeshBufferCount() - 1;
                w->strength = bw.weight;
                w->vertex_id = bw.idVertex;
            }
            
            // skinJoint
            if (bone.weights.size())
            {
                auto boneMat = joint->GlobalMatrix * bone.mtxOffset;
                const unsigned bufferId = mesh->getMeshBufferCount() - 1;

                for (auto iweight = bone.weights.begin(); iweight != bone.weights.end(); iweight ++)
                {
                    auto& weight = (*iweight);
                    const unsigned vertexId = weight.idVertex;

                    auto sourcePos = mesh->getMeshBuffer(bufferId)->getPosition(vertexId);
                    auto sourceNorm = mesh->getMeshBuffer(bufferId)->getNormal(vertexId);
                    irr::core::vector3df destPos, destNormal;
                    boneMat.transformVect(destPos, sourcePos);
                    boneMat.rotateVect(destNormal, sourceNorm);

                    skinnedVertex[vertexId].Moved = true;
                    skinnedVertex[vertexId].Position += destPos * weight.weight;
                    skinnedVertex[vertexId].Normal += destNormal * weight.weight;
                }
            }
        }

        for (size_t i = 0; i < buffer->getVertexCount(); ++i)
        {
            if (skinnedVertex[i].Moved)
            {
                buffer->getPosition(i) = skinnedVertex[i].Position;
                buffer->getNormal(i) = skinnedVertex[i].Normal;
            }
        }
        skinnedVertex.clear();
    }

    // create animation
    for(auto ia = pid->animations.begin(); ia != pid->animations.end(); ia ++) {
        auto& anim = (*ia);
        mesh->setAnimationSpeed(anim.speed);
        for(auto ic = anim.channels.begin(); ic != anim.channels.end(); ic ++) {
            auto& channel = *ic;
            auto joint = findJoint(channel.nodeName.c_str(), mesh);

            for(auto ik = channel.posKeys.begin(); ik != channel.posKeys.end(); ik ++) {
                auto irrKey = mesh->addPositionKey(joint);
                irrKey->frame = (*ik).frame;
                irrKey->position = (*ik).val;
            }
            for(auto ik = channel.rotationKeys.begin(); ik != channel.rotationKeys.end(); ik ++) {
                auto irrKey = mesh->addRotationKey(joint);
                irrKey->frame = (*ik).frame;
                irrKey->rotation = (*ik).val;
            }
            for(auto ik = channel.scaleKeys.begin(); ik != channel.scaleKeys.end(); ik ++) {
                auto irrKey = mesh->addScaleKey(joint);
                irrKey->frame = (*ik).frame;
                irrKey->scale = (*ik).val;
            }
        }
    }
    mesh->setDirty();
    mesh->finalize();
    return mesh;
}

// Wrapper for irr::io::IReadFile

class IrrFileArchive : public BinaryArchive {
public:
    virtual size_t serialize(void* pv, size_t len);
    IrrFileArchive(irr::io::IReadFile* file) { m_file = file; m_bLoading = true; }
protected:
    irr::io::IReadFile* m_file;
};

size_t IrrFileArchive::serialize(void* pv, size_t len) {
    return m_file->read(pv, len);
}

// ImbLoader implementation

ImbLoader::ImbLoader(irr::scene::ISceneManager* smgr, const char* szTxDir) {
    m_smgr = smgr;
    m_szTxDir = szTxDir;
}

ImbLoader::~ImbLoader() {}

irr::scene::IAnimatedMesh* ImbLoader::createMesh(irr::io::IReadFile* file) {
    IrrFileArchive ifa(file);
    SIrrMeshBuild sid;
    if(!serialize(&sid, &ifa)) return NULL;

    for(auto i = m_modifiers.begin(); i != m_modifiers.end(); i ++)
        (*i)->onLoadImbCompleted(file->getFileName(), &sid);

    auto fs = m_smgr->getFileSystem();
    auto fileDir = fs->getFileDir(file->getFileName());

    std::vector<const char*> txDirs;
    txDirs.push_back(m_szTxDir);
    txDirs.push_back(fileDir.c_str());
    txDirs.push_back(NULL);

    auto r = getMeshFromIMB(&sid, m_smgr, txDirs.data());
    for(auto i = m_modifiers.begin(); i != m_modifiers.end(); i ++)
        (*i)->onBuildMeshCompleted(file->getFileName(), r);

    return r;
}

bool ImbLoader::isALoadableFileExtension(const irr::io::path& filename) const {
    auto sz = filename.c_str();
    auto pdot = strrchr(sz, '.');
    return pdot && !_stricmp(pdot + 1, "imb");
}

void ImbLoader::addMeshModifier(ImbMeshModifier* pModifier) {
    m_modifiers.push_back(pModifier);
}
