/*
    imbView.cpp
    Mesh viewer example by Irrlicht engine

    MUTATE Systems 2018 (https://mutate.jp)
*/

#include <Windows.h>
#include <iostream>
#include <irrlicht.h>
#include <vector>
#include "ImbLoader.h"

#pragma comment(lib, "Irrlicht.lib")

using std::vector;
using irr::core::vector3df;
using irr::video::SColor;
using irr::SEvent;
using irr::scene::ILightSceneNode;

class ViewerState : public irr::IEventReceiver
{
    vector3df m_rot;
    bool m_bShadow;
    float m_wheel;
    SEvent::SMouseInput m_miOld;
public:
	ViewerState(void) : m_bShadow(false), m_wheel(0) {
		m_rot.set(5, 45, 0);
        memset(&m_miOld, 0, sizeof(m_miOld));
	}
	vector3df getRotation(void) const { return m_rot; }
    bool isShadowEnable(void) { return m_bShadow; }
    float getW(void) { return m_wheel; }
	virtual bool OnEvent(const SEvent& event);
};

bool ViewerState::OnEvent(const SEvent& event) {
	float rotSpeed = 0.5f;
	switch(event.EventType) {

	case irr::EET_MOUSE_INPUT_EVENT:
		if(m_miOld.isLeftPressed() && event.MouseInput.isLeftPressed() )  {
			m_rot.Y -= (event.MouseInput.X - m_miOld.X) * rotSpeed;
			m_rot.X += (event.MouseInput.Y - m_miOld.Y) * rotSpeed;
		}
        m_wheel += event.MouseInput.Wheel;
		m_miOld = event.MouseInput;
		break;

	case irr::EET_KEY_INPUT_EVENT:
        if(event.KeyInput.PressedDown) {
    		switch(event.KeyInput.Key) {
            case irr::KEY_KEY_S:
                m_bShadow = !m_bShadow;
                break;
            }
        }
		break;
	}
    return false;
}


const wchar_t* APPNAME = L"imbView";

int WINAPI wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
        LPWSTR lpCmdLine, int nCmdShow) {

    if(__argc < 2) {
        ::MessageBoxW(NULL, L"Please specify mesh file name", APPNAME, MB_OK);
        return 1;
    }

    CoInitialize(NULL);

	ViewerState state;

    irr::SIrrlichtCreationParameters creationParams;
//    creationParams.DriverType = irr::video::EDT_OPENGL; // 5
    creationParams.DriverType = irr::video::EDT_DIRECT3D9; // 4
    creationParams.WindowSize = irr::core::dimension2du(1280, 720);
    creationParams.Bits = 32;
    creationParams.Stencilbuffer = true;
    creationParams.Fullscreen = 0;
    creationParams.EventReceiver = &state;
    creationParams.Vsync = true;
    creationParams.AntiAlias = 2;
    auto irrDevice = createDeviceEx(creationParams);
    auto guienv = irrDevice->getGUIEnvironment();

    vector<wchar_t> wszBuf(_MAX_PATH * 2);
    swprintf_s(wszBuf.data(), wszBuf.size(), L"ImbView %s", __wargv[1]);
    irrDevice->setWindowCaption(wszBuf.data());

    auto smgr = irrDevice->getSceneManager();
    ImbLoader imbLoader(smgr, "");
    smgr->addExternalMeshLoader(&imbLoader);

    // Light
    vector3df dirLight(60, 30, 0);

    ILightSceneNode* pLight = smgr->addLightSceneNode();
    pLight->setLightType(irr::video::ELT_DIRECTIONAL);
    irr::video::SLight & l = pLight->getLightData();
    l.AmbientColor = irr::video::SColorf(.2f, .2f, .2f);
    l.DiffuseColor = irr::video::SColorf(0.8f,0.8f,0.8f);
    pLight->setRotation(dirLight);
    pLight->enableCastShadow(false);

    {   // Dummy light for shadows
        auto m_pLightShadow = smgr->addLightSceneNode();
        m_pLightShadow->setRadius(20000);
        m_pLightShadow->setLightType(irr::video::ELT_POINT);
        m_pLightShadow->setPosition(irr::core::vector3df(-3000,10000,-6000));
        m_pLightShadow->enableCastShadow(true);
        irr::video::SLight & l = m_pLightShadow->getLightData();
        l.DiffuseColor = irr::video::SColorf(0, 0, 0);
    }

    auto mesh = smgr->getMesh(__wargv[1]);
	if (!mesh)
	{
		irrDevice->drop();
        ::MessageBoxW(NULL, L"Cannot open mesh file!", APPNAME, MB_OK);
        return 1;
	}

    auto node = smgr->addAnimatedMeshSceneNode( mesh );
    auto psv = node->addShadowVolumeSceneNode(0, -1, true);

    float radius = mesh->getBoundingBox().getExtent().getLength() / 2;

    vector3df camOrg(0, radius/4, radius * 1.2f);

	auto nodeCam = smgr->addCameraSceneNode(0, camOrg, node->getPosition());
    psv->setVisible(false);

	while(irrDevice->run())
	{
        swprintf_s(wszBuf.data(), wszBuf.size(), L"[S] Volume shadow (Current : %s)", state.isShadowEnable() ? L"ON" : L"OFF");

        auto pText = guienv->addStaticText(wszBuf.data(), irr::core::recti(10,10,260,22), true);

        nodeCam->setPosition(camOrg * pow(0.95f, state.getW()));
        dirLight.Y += 1;
        node->setRotation(state.getRotation());
        psv->setVisible(state.isShadowEnable());

        irrDevice->getVideoDriver()->beginScene(true, true, SColor(255,100,101,140));
        smgr->drawAll();
        guienv->drawAll();
		irrDevice->getVideoDriver()->endScene();

        pText->remove();
	}

	irrDevice->drop();
    return 0;
}
